#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <math.h>

#include <hpic.h>

void hpic2ascii_mapheader(char *filename, size_t nside, int order, int coord,
                          char *creator, char *extname, int type)
{
  printf("# File:         %s\n", filename);
  printf("# Creator:      %s\n", creator);
  printf("# ExtName:      %s\n", extname);
  printf("# Nside:        %lu\n", nside);
  if (order == HPIC_RING) {
    printf("# Ordering:     RING\n");
  } else {
    printf("# Ordering:     NEST\n");
  }
  switch (coord) {
  case HPIC_COORD_G:
    printf("# Coordinates:  Galactic\n");
    break;
  case HPIC_COORD_C:
    printf("# Coordinates:  Celestial / Equatorial\n");
    break;
  case HPIC_COORD_E:
    printf("# Coordinates:  Ecliptic\n");
    break;
  default:
    printf("# Coordinates:  Other\n");
    break;
  }
  if (type == HPIC_FITS_FULL) {
    printf("# File Type:    FULL-SPHERE (implicit)\n");
  } else {
    printf("# File Type:    CUT-SPHERE (explicit)\n");
  }
  printf("# -------- Optional Keywords --------\n");
  return;
}

int main(int argc, char *argv[])
{
  char ch = '-';
  size_t i, j, k;

  char infile[HPIC_STRNL];

  size_t nside;
  int order;
  int coord;
  int filetype;
  int tabtype;
  size_t nmaps;
  size_t length;
  size_t nvecs;
  size_t npix;
  int skip;

  char command[HPIC_STRNL];
  char creator[HPIC_STRNL];
  char extname[HPIC_STRNL];

  hpic_vec_int *indx;
  char **vecnames;
  char **vecunits;
  hpic_vec_fltarr *vecs;
  hpic_vec_float *tempvec;

  hpic_int *pix = NULL;
  hpic_int *hits = NULL;
  hpic_float *errs = NULL;
  hpic_keys *keys = NULL;
  hpic_fltarr *maps = NULL;
  hpic_float *tempmap = NULL;

  float val;

  if (argc == 1) {
    printf("\n");
    printf("<******************************************************>\n");
    printf("Usage:  hpic_2ascii <input FITS file>\n");
    printf("<******************************************************>\n\n");
    exit(1);
  }

  /* parse command line */

  strncpy(infile, argv[argc - 1], HPIC_STRNL);

  /* process FITS file */

  if (hpic_fits_map_test(infile, &nside, &order, &coord, &filetype, &nmaps)) {
    keys = hpic_keys_alloc();
    maps = hpic_fltarr_alloc(nmaps);
    for (i = 0; i < nmaps; i++) {
      tempmap = hpic_float_alloc(nside, order, coord, HPIC_AUTO);
      hpic_fltarr_set(maps, i, tempmap);
    }

    if (filetype == HPIC_FITS_FULL) {
      hpic_fits_full_read(infile, creator, extname, maps, keys);
      hpic2ascii_mapheader(infile, nside, order, coord, creator, extname,
                           filetype);
      hpic_keys_printf(keys);
      printf("#\n");
      printf("# ");
      for (i = 0; i < nmaps; i++) {
        tempmap = hpic_fltarr_get(maps, i);
        printf("%10s ", hpic_float_name_get(tempmap));
      }
      printf("\n# ");
      for (i = 0; i < nmaps; i++) {
        tempmap = hpic_fltarr_get(maps, i);
        printf("%10s ", hpic_float_units_get(tempmap));
      }
      printf("\n#\n");
      tempmap = hpic_fltarr_get(maps, 0);
      npix = hpic_float_npix_get(tempmap);
      for (i = 0; i < npix; i++) {
        for (j = 0; j < nmaps; j++) {
          tempmap = hpic_fltarr_get(maps, j);
          printf("%e ", hpic_float_get(tempmap, i));
        }
        printf("\n");
      }
    } else {
      errs = hpic_float_alloc(nside, order, coord, HPIC_AUTO);
      hits = hpic_int_alloc(nside, order, coord, HPIC_AUTO);
      pix = hpic_int_alloc(nside, order, coord, HPIC_AUTO);
      hpic_fits_cut_read(infile, creator, extname, pix, hits, errs, maps,
                         keys);
      hpic2ascii_mapheader(infile, nside, order, coord, creator, extname,
                           filetype);
      hpic_keys_printf(keys);
      printf("#\n");
      printf("# ");
      printf("    PIXELS ");
      for (i = 0; i < nmaps; i++) {
        tempmap = hpic_fltarr_get(maps, i);
        printf("%10s ", hpic_float_name_get(tempmap));
      }
      printf("      HITS       ERRS ");
      printf("\n# ");
      printf("      NONE ");
      for (i = 0; i < nmaps; i++) {
        tempmap = hpic_fltarr_get(maps, i);
        printf("%10s ", hpic_float_units_get(tempmap));
      }
      printf("      NONE ");
      printf("%10s ", hpic_float_units_get(errs));
      printf("\n#\n");
      tempmap = hpic_fltarr_get(maps, 0);
      npix = hpic_float_npix_get(tempmap);
      for (i = 0; i < npix; i++) {
        if (!hpic_is_inull(hpic_int_get(pix, i))) {
          printf("%d ", hpic_int_get(pix, i));
          for (j = 0; j < nmaps; j++) {
            tempmap = hpic_fltarr_get(maps, j);
            printf("%e ", hpic_float_get(tempmap, i));
          }
          printf("%d ", hpic_int_get(hits, i));
          printf("%e ", hpic_float_get(errs, i));
          printf("\n");
        }
      }
      hpic_int_free(pix);
      hpic_int_free(hits);
      hpic_float_free(errs);
    }

    hpic_keys_free(keys);
    for (i = 0; i < nmaps; i++) {
      tempmap = hpic_fltarr_get(maps, i);
      hpic_float_free(tempmap);
    }
    hpic_fltarr_free(maps);
  } else if (hpic_fits_vec_test(infile, &nvecs, &length, &filetype, &tabtype)) {
    keys = hpic_keys_alloc();
    vecnames = hpic_strarr_alloc(nvecs);
    vecunits = hpic_strarr_alloc(nvecs);
    vecs = hpic_vec_fltarr_alloc(nvecs);

    for (i = 0; i < nvecs; i++) {
      tempvec = hpic_vec_float_alloc(length);
      hpic_vec_fltarr_set(vecs, i, tempvec);
    }
    printf("# File:  %s\n", infile);
    if (filetype == HPIC_FITS_VEC) {

      hpic_fits_vec_read(infile, creator, extname, vecs, vecnames, vecunits,
                         keys);
      hpic_keys_printf(keys);
      printf("#\n");
      printf("# ");
      for (i = 0; i < nvecs; i++) {
        printf("%s ", vecnames[i]);
      }
      printf("\n# ");
      for (i = 0; i < nvecs; i++) {
        printf("%s ", vecunits[i]);
      }
      printf("\n#\n");
      for (i = 0; i < length; i++) {
        for (j = 0; j < nvecs; j++) {
          tempvec = hpic_vec_fltarr_get(vecs, j);
          printf("%e ", hpic_vec_float_get(tempvec, i));
        }
        printf("\n");
      }
    } else {
      indx = hpic_vec_int_alloc(length);
      hpic_fits_vecindx_read(infile, creator, extname, indx, vecs, vecnames,
                             vecunits, keys);
      hpic_keys_printf(keys);
      printf("#\n");
      printf("# INDEX ");
      for (i = 0; i < nvecs; i++) {
        printf("%s ", vecnames[i]);
      }
      printf("\n# NONE ");
      for (i = 0; i < nvecs; i++) {
        printf("%s ", vecunits[i]);
      }
      printf("\n#\n");
      for (i = 0; i < length; i++) {
        printf("%d ", hpic_vec_int_get(indx, i));
        for (j = 0; j < nvecs; j++) {
          tempvec = hpic_vec_fltarr_get(vecs, j);
          printf("%e ", hpic_vec_float_get(tempvec, i));
        }
        printf("\n");
      }
      hpic_vec_int_free(indx);
    }

    hpic_keys_free(keys);
    hpic_strarr_free(vecnames, nvecs);
    hpic_strarr_free(vecunits, nvecs);
    for (i = 0; i < nvecs; i++) {
      tempvec = hpic_vec_fltarr_get(vecs, i);
      hpic_vec_float_free(tempvec);
    }
    hpic_vec_fltarr_free(vecs);

  } else {
    fprintf(stderr, "hpic_2ascii:  FITS file is not a recognized type\n");
    exit(1);
  }

  return 0;
}
